from threading import Thread
import time
import Queue
import random
import atexit

import dns.resolver
import dns.query
import dns.exception

import storage
import log

class VerificationThread(Thread):

    rootservers = {'A.ROOT-SERVERS.NET.':'198.41.0.4',
                   'B.ROOT-SERVERS.NET.':'192.228.79.201',
                   'C.ROOT-SERVERS.NET.':'192.33.4.12',
                   'D.ROOT-SERVERS.NET.':'128.8.10.90',
                   'E.ROOT-SERVERS.NET.':'192.203.230.10',
                   'F.ROOT-SERVERS.NET.':'192.5.5.241',
                   'G.ROOT-SERVERS.NET.':'192.112.36.4',
                   'H.ROOT-SERVERS.NET.':'128.63.2.53',
                   'I.ROOT-SERVERS.NET.':'192.36.148.17',
                   'J.ROOT-SERVERS.NET.':'192.58.128.30',
                   }

    def __init__(self, server):
        self.pool = server.pool
        self.vcache = server.verifiedcache
        self.alerter = server.alerter
        self._running = False
        self.nscache = {}
        self.peers = server.config.get('global','peers').split(',')
        atexit.register(self.stop)
        Thread.__init__(self)
        
    def verify(self, recset):
        # Verify against current
        pg = self.vcache.get_peergroup(recset.name, recset.type)
        if pg:
            ret = pg.verify(recset)
            if ret:
                return True
        # Verify against auth
        ret = self.verify_auth_answers(recset)
        if ret:
            return True
        pg = self.vcache.get_peergroup(recset.name, recset.type)
        # Verify against peers
        ret = self.verify_peer_answers(recset)
        if ret:
            return True
        # Is it a cdn? fast flux?
        if not pg:
            pg = self.vcache.get_peergroup(recset.name, recset.type)
            if not pg:
                self.alerter.error(recset, 'Validation Failed')
                return False
        if pg.is_cdn():
            if pg._verify_cdn(recset):
                return True
            else:
                self.alerter.warning(recset, 'CDN Detected')
        else:
            self.alerter.error(recset, 'Validation Failed')
        return False
    
    def start(self):
        self._running = True
        Thread.start(self)
   
    def stop(self):
        self._running = False
              
    def run(self):
        while self._running:
            try:
                recset = self.pool.get_nowait()
            except Queue.Empty:
                time.sleep(.5)
            else:
                try:
                    self.verify(recset)
                except Exception:
                    log.tb()
                    print "Exception Verifying", recset

    # NEW VERIFICATION LOGIC (START)

    def add_to_nscache(self, name, ttl, nameservers):
        maxts = time.time() + ttl
        name = name.lower()
        self.nscache[name] = (maxts, nameservers)

    def get_from_nscache(self, name):
        name = name.lower()
        if name == '':
            return self.rootservers
        try:
            maxts,nameservers = self.nscache[name]
            if maxts < time.time():
                return None
        except KeyError:
            return None
            
    def get_closest_nameservers(self, name):
        nameservers = self.rootservers
        if type == 'NS':
            lname = '.'.join(name.split('.')[1:])
        else:
            lname = name
        lparts = lname.split('.')
        for i in range(1,len(lparts)):
            ln = '.'.join(lparts[i:])
            # Check Cache for name of 
            nameservers = self.get_from_nscache(ln)
        return nameservers
   
    def get_response_from_nameservers(self, vrecset, nsname, nsip):
        # TODO: FiX This only queries the first nameserver first        
        #print "dig @%s %s %s" %(nsip, vrecset.name, vrecset.type)
        if not nsip:
            recset = storage.RecordSet(None, nsname, 'A')
            rs = self.get_answers(recset)
            if rs:
                nsip = rs.records[0].data
            else: 
                raise dns.exception.Timeout
            #print "dig @%s %s %s" %(nsip, vrecset.name, vrecset.type)
        query = dns.message.make_query(vrecset.name, vrecset.type, 1)
        response = dns.query.udp(query, nsip, 1)
        return response

    def get_answers(self, vrecset):
        nameservers = self.get_closest_nameservers(vrecset.name)
        # What are the max labels in a domain name.
        # Only follow a max of 25 recursions 
        for i in range(5): 
            nitems = nameservers.items()
            ri = random.randint(0,len(nitems)-1)
            nslist = nitems[ri:]
            nslist.extend(nitems[:ri])
            #print " checking",nslist
            for nsname,nsip in nslist:
                response = None
                for i in range(3):
                    try:
                        response = self.get_response_from_nameservers(vrecset, nsname, nsip)
                        break
                    except dns.exception.Timeout:
                        continue
                    except Exception:
                        continue
                if not response:
                    #print "Too Many Timeouts/Exceptions for query: dig @%s %s %s" % (nsip, vrecset.name, vrecset.type)
                    continue
                if response.answer:
                    # Compare the results
                    answer = response.answer[0]
                    rs = storage.RecordSet(None, vrecset.name, vrecset.type)
                    for item in answer.items:
                        rs.add_record(item.to_text(), 0)
                    return rs
                if response.authority:
                    # if the 
                    rrset = response.authority[0]
                    if rrset.name.to_text() == '.':
                        print "Deferred to the roots! Bad Delegation %s %s" % (vrecset.name, vrecset.type)
                        return None
                    if rrset.rdtype == 2:
                        nameservers = {}
                        for rr in rrset.items:
                            nameservers[str(rr.to_text())] = None
                        if response.additional:
                            for rr in response.additional:
                                if rr.rdtype == 1:
                                    if str(rr.name) in nameservers.keys():
                                        nameservers[str(rr.name)] = rr.items[0].to_text()
                        self.add_to_nscache(rrset.name.to_text(), rrset.ttl, nameservers)
                        break

 
    def verify_auth_answers(self, vrecset):
        oldns = {}
        nameservers = self.get_closest_nameservers(vrecset.name)
        # What are the max labels in a domain name.
        validated_ns = False
        # Only follow a max of 25 recursions 
        for i in range(25): 
            if oldns == nameservers:
                break
            else:
                oldns = nameservers
            nitems = nameservers.items()
            ri = random.randint(0,len(nitems)-1)
            nslist = nitems[ri:]
            nslist.extend(nitems[:ri])
            #print " checking",nslist
            for nsname,nsip in nslist:
                response = None
                for i in range(3):
                    try:
                        response = self.get_response_from_nameservers(vrecset, nsname, nsip)
                        break
                    except dns.exception.Timeout:
                        continue
                if not response:
                    #print "Too Many Timeouts for query: dig @%s %s %s" % (nsip, vrecset.name, vrecset.type)
                    continue
                if response.answer:
                    # Compare the results
                    answer = response.answer[0]
                    rs = storage.RecordSet(None, vrecset.name, vrecset.type)
                    for item in answer.items:
                        rs.add_record(item.to_text(), 0)
                    self.vcache.add_auth_recordset(nsname,vrecset.name,vrecset.type,rs)
                    if vrecset == rs:
                        return True
                    continue
                if response.authority:
                    # if the 
                    rrset = response.authority[0]
                    if rrset.name.to_text() == '.':
                        if validated_ns:
                            return True
                        return False
                    if rrset.rdtype == 2:
                        if rrset.name.to_text().lower() == vrecset.name and vrecset.type == 'NS':
                            # Some Nameservers do not have the ns records in the zone
                            rs = storage.RecordSet(None, vrecset.name, vrecset.type)
                            for item in rrset.items:
                                rs.add_record(item.to_text(), 0)
                            #self.vcache.add_auth_recordset(nsname,vrecset.name,vrecset.type,rs)
                            if vrecset == rs:
                                validated_ns = True
                        nameservers = {}
                        for rr in rrset.items:
                            nameservers[str(rr.to_text())] = None
                        if response.additional:
                            for rr in response.additional:
                                if rr.rdtype == 1:
                                    if str(rr.name) in nameservers.keys():
                                        nameservers[str(rr.name)] = rr.items[0].to_text()
                        self.add_to_nscache(rrset.name.to_text(), rrset.ttl, nameservers)
                        break
        if validated_ns:
            # This case comes up when their nameservers are timing out 
            return True    
        return False
                

        
    # NEW VERIFICATION LOGIC (END)
    def DELME_verify_auth_answers(self, vrecset):
        lname = vrecset.name
        type = vrecset.type
        #print "verify_auth_answers(",lname, type,")"
        authns = self.rootservers
        delegs =  self.delegations.keys()
        lp = lname.split('.')
        for i in range(0, len(lp)):
            idx = i+1
            n = '.'.join(lp[-idx:])
            n = n.lower()
            if n in delegs:
                authns = self.delegations[n]
        offset = random.randint(0,len(self.rootservers)-1)
        try:
            authrecords = self._query_ns(lname, type, authns)
        except:
            authrecords = self._query_ns(lname, type, authns)

        for nsname in authrecords.keys():
            nsip = authrecords[nsname]
            response = self._query(lname, type, nsname=nsname, nsip=nsip, 
                                   auth=authrecords)
            if response.answer:
                answer = response.answer[0]
                rs = storage.RecordSet(None, lname, type)
                for item in answer.items:
                    rs.add_record(item.to_text(), 0)
                self.vcache.add_auth_recordset(nsname,lname,type,rs)
                if vrecset == rs:
                    return True
            if type == 'NS' and response.authority:
                answer = response.authority[0]
                rs = storage.RecordSet(None, lname, type)
                for item in answer.items:
                    rs.add_record(item.to_text(), 0)
                self.vcache.add_auth_recordset(nsname,lname,type,rs)
                if vrecset == rs:
                    return True
        return False

    def verify_peer_answers(self, vrecset):
        qname = vrecset.name
        type = vrecset.type
        #print "verify_auth_answers(%s, %s)" % (qname, type)
        query = dns.message.make_query(qname, type, 1)
        for resolver in self.peers:
            try:
                response = dns.query.udp(query, resolver, 1)
            except dns.exception.Timeout:
                continue
            if response.answer:
                answer = response.answer[0]
                rs = storage.RecordSet(None, qname, type)
                for item in answer.items:
                    rs.add_record(item.to_text(), 0)
                self.vcache.add_peer_recordset(resolver, qname, type, rs)
        pg = self.vcache.get_peergroup(qname, type)
        if not pg:
            return False
        return pg._validate_peer(vrecset, 2)

    def _find_record(self):
        return

    cache = {}
    delegations = {}

    def query_dns(self,query, ip):
        for i in range(3):
            try:
                response = dns.query.udp(query, ip, 1)
            except dns.exception.Timeout:
                pass
            else:
                return response
        raise

        
    def _query(self, lname, type, nsname=None, nsip=None, auth={}):
        #print "_query(",lname, type, nsname, nsip, auth, ")"
        try:
            response = self.cache[lname][type]
            return response
        except KeyError:
            pass
        if not nsip:
            ret = self.query_record(nsname, 'A')
            recs = ret.get_records()
            offset = random.randint(0,len(recs)-1) 
            nsip = recs[offset].data
            auth[nsname] = nsip
        #print "  dig @%s %s %s" % (nsip, lname, type)
        query = dns.message.make_query(lname, type, 1)
        response = self.query_dns(query, nsip)
        # Populate cache
        if not response.answer:
            return response
        if not self.cache.has_key(lname):
            self.cache[lname] = {}
        self.cache[lname][type] = response
        # Return
        return response

    def _query_ns(self, lname, type, authrecords):
        #print "_query_ns(",lname,type,authrecords,")"
        # get ip to query from authrecords
        # Perferably grab glued records... Meh
        values = authrecords.values()
        ips = []
        for value in values:
            if value:
                ips.append(value)
        if ips:
            offset = random.randint(0,len(ips)-1) 
            nsname = None
            nsip = ips[offset]
        else:
            offset = random.randint(0,len(authrecords.keys())-1) 
            nsname = authrecords.keys()[offset]
            nsip = authrecords[nsname]
        # dig @nameserverip lname SOA ... (was querying for NS)
        if type == 'NS':
            qtype = 'NS'
        else:
            qtype='SOA'
        response = self._query(lname, qtype, nsname=nsname, nsip=nsip)
        # if retuned auth is SOA:
        if response.answer:
            return authrecords
        if not response.authority and qtype=='SOA':
            return authrecords
        if response.authority[0].rdtype == 6:
            # verified this is the authset
            return authrecords
        # If returned AUTHORITY is NS:
        authority = response.authority
        rrset = response.authority[0]
        if rrset.name.to_text() == '.':
            raise "Deferred to Roots... Bad Delegation"
        if qtype == 'NS' and rrset.name.to_text() == lname:
            return authrecords
        
        if rrset.rdtype == 2:
            authrecords = {}
            for rr in rrset.items:
                authrecords[str(rr.to_text())] = None
            for rr in response.additional:
                if rr.rdtype == 1:
                    if str(rr.name) in authrecords.keys():
                        authrecords[str(rr.name)] = rr.items[0].to_text()
            self.delegations[str(rrset.name.to_text()).lower()] = authrecords
            return self._query_ns(lname, type, authrecords)
        return None

    def query_record(self, lname, type):
        #print "query_record(",lname,type,")"
        # Find authoritative ns
        authns = self.rootservers
        delegs = self.delegations.keys()
        #print " PoolSize:",self.pool.qsize() 
        lp = lname.split('.')
        for i in range(0, len(lp)):
            idx = i+1
            n = '.'.join(lp[-idx:])
            n = n.lower()
            if n in delegs:
                authns = self.delegations[n]
                #print " Found Delegations", authns
        offset = random.randint(0,len(self.rootservers)-1)
        authrecords = self._query_ns(lname, type, authns)
        # who will you query?
        offset = random.randint(0,len(authrecords.keys())-1)
        nsname = authrecords.keys()[offset]
        nsip = authrecords[nsname]
        # dig @nsip lname type
        response = self._query(lname, type, nsname=nsname, nsip=nsip)
        if response.answer:
            answer = response.answer[0]
            rs = storage.RecordSet(None, lname, type)
            for item in answer.items:
                rs.add_record(item.to_text(), 0)
            return rs
        return None
