'''
This stores the cache data
'''
import time

class Record(object):

    def __init__(self, recordset, data, ttl, trustfactor=None):
        self.recordset = recordset
        self.data = data
        self.ttl = ttl
        self.trustfactor = trustfactor

    def __cmp__(self, rec):
        if self.recordset.type == 'NS':
            mydata = self.data.lower()
            rcdata = rec.data.lower()
        else:
            mydata = self.data
            rcdata = rec.data
        if mydata != rcdata:
            return 1
        else:
            return 0
    
    def __str__(self):
        return '<Record (\'%s\')>' % self.data


class RecordSet(object):
    
    def __init__(self, parent, name, type):
        self.name = name
        self.type = type
        self.storage = parent
        self.server = None
        self.records = []
        return

    def add_record(self, data, ttl):
        rec = Record(self, data, ttl)
        self.records.append(rec)
        return True

    def get_records(self):
        return self.records

    def __cmp__(self, rset):
        validated = 0
        if len(rset.records) != len(self.records):
            return 1
        for r in rset.records:
            # Validate this record against the current recordset
            for record in self.records:
                if r == record:
                    validated += 1
                    break
        if validated == len(rset.records):
            return 0
        return 1

    def __str__(self):
        data = []
        for rec in self.records: 
            data.append(rec.data)
        return '<RecordSet  %s %s (\'%s\')>' % (self.name, self.type,
            '\',\''.join(data))


class CacheStorage(object):

    def __init__(self, timeOfDump=None):
        if timeOfDump:
            self.time = timeOfDump
        else:
            self.time = time.time()
        self.records = {}
   
    def __set_structure(self, name, type):
        if not self.records.has_key(name):
            self.records[name] = {}
        if not self.records[name].has_key(type):
            self.records[name][type] = RecordSet(self, name, type)

    def add_record(self, name, type, data, ttl):
        if not name.endswith('.'):
            name += '.'
        self.__set_structure(name, type)
        self.records[name][type].add_record(data,ttl)

    def get_recordset(self, name, type):
        try:
            recordset = self.records[name][type]
        except KeyError:
            return None
        return recordset

    def diff_with_old(self, oldcache):
        diff_recs = []
        newct = -1
        chgct = 0
        for name in self.records:
            for type in self.records[name]:
                orecordset = oldcache.get_recordset(name, type)
                vrecordset = self.get_recordset(name, type)
                if not orecordset:
                    newct += 1
                    diff_recs.append(vrecordset)
                elif orecordset != vrecordset:
                    chgct += 1
                    diff_recs.append(vrecordset)
        print "STATS: New: %d Changed: %d" % (newct, chgct)
        return diff_recs


class PeerGroup(object):
    
    def __init__(self, parent, name, type):
        self.storage = parent
        self.name = name
        self.type = type
        self.peers = {}
        self.auths = {}
        self.history = []
        return

    def add_peer_recordset(self, peer, recordset):
        if peer in self.storage.validpeers:
            raise errors.InvalidPeer
        if not self.peers.has_key(peer):
            self.peers[peer] = recordset
        else:
            orecset = self.peers[peer]
            # Currently we only care about the history of auth
            self.peers[peer] = recordset

    def get_peer_recordset(self, peer):
        return self.peers[peer]

    def add_auth_recordset(self, auth, recordset):
        if not self.auths.has_key(auth):
            self.auths[auth] = recordset
        else:
            orecset = self.auths[auth]
            if orecset not in self.history:
                self.history.append(orecset)
            self.auths[auth] = recordset

    def get_auth_recordset(self, auth):
        return self.auths[auth]

    def _validate_auth(self, vrecset): 
        for recset in self.auths.values():
            if vrecset == recset:
                # The authoritatives knew whom this was
                return True
        return False

    def _validate_peer(self, vrecset, threshold):
        weight = 0
        for recset in self.peers.values():
            if vrecset == recset:
                weight += 1
        if weight >= threshold:
            return True
        return False

    def _verify(self, vrecset, peerdepth):
        if self._validate_auth(vrecset):
            return True
        if self._validate_peer(vrecset, peerdepth):
            return True
        return False

    def verify(self, vrecset, peerdepth=2):
        return self._verify(vrecset, peerdepth)

    def is_fastflux(self):
        return False

    def is_cdn(self):
        # if # of auth variations > 2: return cdn
        uniques = []
        for auth in self.auths.keys():
            recset = self.get_auth_recordset(auth)
            seen = False
            for u in uniques:
                if recset == u:
                    seen = True
                    break
            if not seen:
                uniques.append(recset)
        if len(uniques) > 2:
            return True
        return False

    def _verify_history(self, vrecset):
        for vrec in vrecset.get_records():
            foundrec = False
            for hrecset in self.history:
                if vrec in hrecset.get_records():
                    foundrec = True
                    continue
            if not foundrec:
                return False
        return True

    def _verify_cdn(self, vrecset):
        '''Here we will do network verification'''
        if self.type != 'A':
            return False
        networks = []
        for ips in self.auths.values():
            for ip in ips.get_records():
                network = '.'.join(ip.data.split('.')[:3])
                if network not in networks:
                    networks.append(network)
        for ip in vrecset.get_records():
            foundrec = False
            for network in networks:
                if ip.data.startswith(network):
                    foundrec = True
                    break
            if not foundrec:
                return False
        print "Validated in CDN!!!"
        return True

    def __str__(self):
        s = "<PeerGroup %s %s Auth {" % (self.name,self.type)
        for auth in self.auths.keys():
            s += "'%s':" % auth
            s += str(self.auths[auth])
            s += ","
        s+= "} Peer {}"
        s+= ">" 
        return s


class VerifiedCache(object):

    def __init__(self, validpeers=[]):
        self.validpeers = validpeers
        self.records = {}

    def __set_structure(self, name, type):
        if not self.records.has_key(name):
            self.records[name] = {}
        if not self.records[name].has_key(type):
            self.records[name][type] = PeerGroup(self, name, type)

    def add_peer_recordset(self, peer, name, type, recordset):
        self.__set_structure(name, type)
        self.records[name][type].add_peer_recordset(peer, recordset)

    def add_auth_recordset(self, peer, name, type, recordset):
        self.__set_structure(name, type)
        self.records[name][type].add_auth_recordset(peer, recordset)

    def get_peergroup(self, name, type):
        try:
            return self.records[name][type]
        except KeyError:
            return None
